import pytest
import scivision.catalog
from pydantic import ValidationError
from pathlib import Path
from scivision.catalog import PandasCatalog, CatalogModels


class TestPandasCatalogInit:
    """Checks that PandasCatalogs can be constructed from several argument
    types."""

    def test_init_default(self):
        """Load the default catalog. Check that no exceptions are raised
        (default catalog may get updated, so no further checks)
        """

        PandasCatalog()

    def test_init_filenames(self):
        """Load catalog from explicit paths, and check that the entries are as
        expected
        """
        cat = PandasCatalog(
            "tests/test_datasource_catalog.json", "tests/test_model_catalog.json"
        )

        models = cat.models.to_dataframe()
        datasources = cat.datasources.to_dataframe()

        assert len(models == 2)
        assert len(datasources == 2)
        assert models.iloc[0]["name"] == "example-model-1"
        assert datasources.iloc[0]["name"] == "example-datasource-1"

    def test_init_bad(self):
        """Check that an exception is raised when the json does not conform to
        the schema
        """
        with pytest.raises(ValidationError):
            PandasCatalog(
                datasources="tests/test_datasource_catalog_bad.json",
                models="tests/test_model_catalog.json",
            )

    def test_init_entry(self):
        """Load catalog from CatalogDatasources/CatalogModels instances"""

        # Instances created from json strings separately first
        datasources = scivision.catalog.CatalogDatasources.parse_raw(
            Path("tests/test_datasource_catalog.json").read_text()
        )

        models = scivision.catalog.CatalogModels.parse_raw(
            Path("tests/test_model_catalog.json").read_text()
        )
        # Then used to initialize the catalog
        cat = PandasCatalog(datasources, models)

        models = cat.models.to_dataframe()
        datasources = cat.datasources.to_dataframe()

        assert len(models == 2)
        assert len(datasources == 2)
        assert models.iloc[0]["name"] == "example-model-1"
        assert datasources.iloc[0]["name"] == "example-datasource-1"


class TestCompatible:
    """Checks that compatible models and datasources in a catalog can be
    correctly identified"""

    @pytest.fixture
    def cat(self):
        return PandasCatalog(
            datasources="tests/test_datasource_catalog.json",
            models="tests/test_model_catalog.json",
        )

    def test_compatible_models_named(self, cat):
        assert len(cat.compatible_models("example-datasource-1").to_dataframe()) == 2
        assert len(cat.compatible_models("example-datasource-2").to_dataframe()) == 1

    def test_compatible_datasources_named(self, cat):
        assert len(cat.compatible_datasources("example-model-1").to_dataframe()) == 1
        assert len(cat.compatible_datasources("example-model-2").to_dataframe()) == 2

    def test_compatible_models_dict(self, cat):
        compat = cat.compatible_models(
            {"labels_provided": False, "tasks": ["object-detection"], "format": "image"}
        ).to_dataframe()
        assert len(compat) == 1 and compat["name"].item() == "example-model-2"

    def test_compatible_datasources_dict(self, cat):
        compat = cat.compatible_datasources(
            {"labels_required": False, "tasks": ["object-detection"], "format": "image"}
        ).to_dataframe()
        assert len(compat) == 1 and compat["name"].item() == "example-datasource-1"


def test_query_to_catalog_entry():
    """Check that a pair of CatalogModels/CatalogDatasources can be
    converted to a PandasCatalog and back
    """

    datasources_cat_expected = scivision.catalog.CatalogDatasources.parse_raw(
        Path("tests/test_datasource_catalog.json").read_text()
    )
    models_cat_expected = scivision.catalog.CatalogModels.parse_raw(
        Path("tests/test_model_catalog.json").read_text()
    )
    cat = PandasCatalog(datasources_cat_expected, models_cat_expected)

    models_cat_actual = CatalogModels(
        entries=cat.models.to_dict(),
        catalog_type=models_cat_expected.catalog_type,
        name=models_cat_expected.name,
    )
    assert models_cat_actual == models_cat_expected



import scivision.catalog


class DatasourceWithExamples(scivision.catalog.CatalogDatasourceEntry):
    class Config:
        @staticmethod
        def schema_extra(schema, model):
            cat = scivision.catalog.default_catalog.datasources.to_dataframe()

            tags = list(cat.tags.explode().unique())
            schema['properties']['tags']['items']['examples'] = tags

            domains = list(cat.domains.explode().unique())
            schema['properties']['domains']['items']['examples'] = domains

            formats = list(cat.format.explode().unique())
            schema['properties']['format']['examples'] = formats

            institutions = list(cat.institution.explode().unique())
            schema['properties']['institution']['items']['examples'] = institutions


def entry_point():
    """
    Write the json schema for datasource and model catalog entries to files

    This is the entry point for the 'scivision-catalog-json-schema' command.

    This command writes the files datasource_schema.js and model_schema.js.
    """

    datasource_schema = DatasourceWithExamples.schema_json()
    with open("datasource_schema.js", "w") as f:
        print(f"const schema = {datasource_schema};", file=f)
        print("export default schema;", file=f)

    model_schema = scivision.catalog.CatalogModelEntry.schema_json()
    with open("model_schema.js", "w") as f:
        print(f"const schema = {model_schema};", file=f)
        print("export default schema;", file=f)



from scivision.io import load_dataset, load_pretrained_model, wrapper, _parse_url, _get_model_configs
import intake
import pytest
import fsspec
import xarray
import yaml


def test_parse_url():
    """Test that GitHub urls are correctly converted to raw."""
    path = 'https://github.com/alan-turing-institute/scivision/tests/test_reader.py'
    raw = 'https://raw.githubusercontent.com/alan-turing-institute/scivision/main/tests/test_reader.py'
    assert _parse_url(path) == raw


def test_parse_url_blob():
    """Test that GitHub urls directly copied from a browser are correctly converted to raw."""
    path = 'https://github.com/alan-turing-institute/scivision/blob/main/tests/test_reader.py'
    raw = 'https://raw.githubusercontent.com/alan-turing-institute/scivision/main/tests/test_reader.py'
    assert _parse_url(path) == raw


def test_parse_url_branch():
    """Test that GitHub urls are correctly converted to raw when a branch is specified."""
    # note: test-branch does not exist
    branch = 'test-branch'
    path = 'https://github.com/alan-turing-institute/scivision/tests/test_reader.py'
    raw = 'https://raw.githubusercontent.com/alan-turing-institute/scivision/' + branch + '/tests/test_reader.py'
    assert _parse_url(path, branch=branch) == raw


def test_load_dataset_remote():
    """Test that an intake catalog is generated from scivision-data.yml file in an example GitHub repo."""
    commit_hash = '473a7fd02f4cdd2dd261208a791c7da76cd413a0'
    assert isinstance(
        load_dataset('https://github.com/alan-turing-institute/intake-plankton', branch=commit_hash),
        intake.catalog.local.YAMLFileCatalog
    )


def test_load_dataset_branch_and_diff_file_name():
    """Test that an intake catalog is generated when specifying a branch AND that a custom file name."""
    branch = 'diff-name-yml'
    assert isinstance(
        load_dataset('https://github.com/alan-turing-institute/intake-plankton/thESciViSionYAMLfileee.yaml', branch=branch),
        intake.catalog.local.YAMLFileCatalog
    )


def test_load_dataset_local():
    """Test that an intake catalog is generated from a local yml and can be converted to xarray."""
    cat = load_dataset('tests/test_dataset_scivision.yml')
    assert isinstance(cat, intake.catalog.local.YAMLFileCatalog)
    assert isinstance(cat.test_images().to_dask(), xarray.core.dataarray.DataArray)


def test_load_dataset_local_zip():
    """Test that an intake catalog is generated from a local yml which loads from a zip file and can be converted to xarray."""
    cat = load_dataset('tests/test_dataset_scivision_zip.yml')
    assert isinstance(cat, intake.catalog.local.YAMLFileCatalog)
    assert isinstance(cat.test_images().to_dask(), xarray.core.dataarray.DataArray)


def test_get_model_configs():
    """Test that a config with multiple models is split into separate configs."""
    path = 'tests/test_multiple_models_scivision.yml'
    file = fsspec.open(path)
    with file as config_file:
        stream = config_file.read()
        config = yaml.safe_load(stream)
    config_list = _get_model_configs(config, load_multiple=True)
    for config in config_list:
        assert 'name' in config
        assert 'model' in config


def test_load_pretrained_model_remote():
    """Test that scivision can load a pretrained model from an example GitHub repo."""
    commit_hash = '77bbe037234d11538e46c3ced3d2a5f9294c8468'
    assert isinstance(
        load_pretrained_model('https://github.com/alan-turing-institute/scivision-test-plugin/.scivision-config_imagenet.yaml', allow_install=True, branch=commit_hash),
        wrapper.PretrainedModel
    )


def test_load_pretrained_model_local():
    """Test that scivision can load a pretrained model from a local yaml that points to a GitHub repo."""
    assert isinstance(
        load_pretrained_model('tests/test_model_scivision.yml', allow_install=True),
        wrapper.PretrainedModel
    )


def test_load_named_pretrained_model_local():
    """Test that scivision can load a specific model from the given scivision.yml."""
    assert isinstance(
        load_pretrained_model('tests/test_model_scivision.yml', allow_install=True, model_selection='ImageNetModel'),
        wrapper.PretrainedModel
    )
    assert isinstance(
        load_pretrained_model('tests/test_multiple_models_scivision.yml', allow_install=True, model_selection='ImageNetModel'),
        wrapper.PretrainedModel
    )


def test_load_wrong_model_name_raises_value_error():
    """Test that a value error is raised when a model name is specified that doesn't match the model in the config."""
    with pytest.raises(ValueError):
        load_pretrained_model('tests/test_model_scivision.yml', allow_install=True, model_selection='FakeModel')


def test_load_wrong_model_name_raises_value_error_config_has_multiple_models():
    """Test that a value error is raised when a model name is specified that doesn't match one of the models in the config."""
    with pytest.raises(ValueError):
        load_pretrained_model('tests/test_multiple_models_scivision.yml', allow_install=True, model_selection='FakeModel')


def test_load_multiple_models():
    """Test that scivision can load multiple pretrained models from the same GitHub repo."""
    model1, model2 = load_pretrained_model('tests/test_multiple_models_scivision.yml', allow_install=True, load_multiple=True)
    assert isinstance(model1, wrapper.PretrainedModel)
    assert isinstance(model2, wrapper.PretrainedModel)


def test_load_first_model_from_config_with_multiple():
    """Test that scivision loads a single a model from a config with multiple models when not told to load multiple."""
    assert isinstance(
        load_pretrained_model('tests/test_multiple_models_scivision.yml', allow_install=True, load_multiple=False),
        wrapper.PretrainedModel
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import pkgutil
import pandas as pd
import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, FrozenSet, Optional, Tuple, Union
from pydantic import AnyUrl, BaseModel, Field, validator
from enum import Enum
from collections import Counter


class TaskEnum(str, Enum):
    classificiation = "classification"
    object_detection = "object-detection"
    segmentation = "segmentation"
    thresholding = "thresholding"
    other = "other"


class FlexibleUrl(AnyUrl):
    host_required = False


class CatalogModelEntry(BaseModel, extra="forbid", title="A model catalog entry"):
    # tasks, institution and tags are Tuples (rather than Lists) so
    # that they are immutable - Tuple is being used as an immutable
    # sequence here. This means that these fields are hashable, which
    # can be more convenient when included in a dataframe
    # (e.g. unique()). Could consider using a Frozenset for these
    # fields instead, since duplicates and ordering should not be
    # significant.
    name: str = Field(
        ...,
        title="Name",
        description="Short, unique name for the model (one or two words, "
        "under 20 characters recommended)",
    )
    description: Optional[str] = Field(
        None,
        title="Description",
        description="Detailed description of the model",
    )
    tasks: Tuple[TaskEnum, ...] = Field(
        (),
        title="Tasks",
        description="Which task (or tasks) does the model perform?",
    )
    url: FlexibleUrl = Field(
        ...,
        title="URL",
        description="The URL path to the scivision yml file in the model repo. "
        "If the exact path is not specified, Scivision will try to locate the file in"
        " the default location at .scivision/model.yml",
    )
    pkg_url: str = Field(
        ...,
        title="Python package",
        description="A pip requirement specifier for PyPI, or a URL of the "
        "archive or package (on GitHub, for exampe)",
    )
    format: str = Field(
        ...,
        title="Model input format",
        description="The type of data consumed by the model",
    )
    scivision_usable: bool = Field(
        False,
        title="Can the model be installed and loaded with the scivision Python package, with "
        "scivision.load_pretrained_model(<model url>, allow_install=True)?"
        "Leave unchecked if not sure",
    )
    pretrained: bool = Field(
        True,
        title="Pretrained model?",
    )
    labels_required: bool = Field(
        True,
        title="Labels required?",
        description="Does the model require labeled data for training?",
    )
    institution: Tuple[str, ...] = Field(
        (),
        title="Institution(s)",
        description="A list of institutions that produced or are associated with "
        "the model (one per item)",
    )
    tags: Tuple[str, ...]

    def __getitem__(self, item):
        return getattr(self, item)


class CatalogModels(BaseModel, extra="forbid"):
    catalog_type: str = "scivision model catalog"
    name: str
    # Tuple: see comment on CatalogModelEntry
    entries: Tuple[CatalogModelEntry, ...]

    @validator("entries")
    def name_unique_key(cls, entries):
        name_counts = Counter([entry['name'] for entry in entries])
        dups = [item for item, count in name_counts.items() if count > 1]

        if dups:
            raise ValueError(f"The 'name' field in the model catalog should be unique (duplicates: {dups})")

        return entries


class CatalogDatasourceEntry(
    BaseModel, extra="forbid", title="Datasource catalog entry"
):
    name: str = Field(
        ...,
        title="Name",
        description="Short name for the datasource, that must be unique among "
        "all catalog entries (one or two words, under 20 characters recommended)",
    )
    description: Optional[str] = Field(
        None,
        title="Description",
        description="Detailed description of the dataset (no length limit)",
    )
    tasks: FrozenSet[TaskEnum] = Field(
        None,
        title="Suitable tasks",
        description="For which task or tasks is this datasource likely to be "
        "suitable? (Select any number of the following items)",
    )
    labels_provided: bool = Field(
        False,
        title="Labels provided",
        description="Is this a labelled dataset? This can make it suitable for training or validation",
    )
    domains: Tuple[str, ...] = Field(
        None,
        title="Domain areas",
        description="Which domain area or areas is this datasource from? (One per item, no duplicates)",
        # Note: using uniqueItems (used for the json schema) rather
        # than unique_items (which is not possible to enforce on a
        # Tuple).  Could use a set/frozenset instead, or a tuple
        # variant with a constraint.
        uniqueItems=True,
    )
    url: FlexibleUrl = Field(
        None,
        title="URL",
        description="The URL path to the scivision yml file in the datasource repo. "
        "If the exact path is not specified, Scivision will try to locate the file in"
        " the default location at .scivision/data.yml",
    )
    format: str = Field(
        None,
        title="Format",
    )
    institution: Tuple[str, ...] = Field(
        (),
        title="Institution(s)",
        description="A list of institutions that produced or are associated with "
        "the dataset (one per item)",
    )
    tags: Tuple[str, ...] = Field(
        (),
        title="Tags",
        description="A list of free-form data to associate with the dataset",
    )

    def __getitem__(self, item):
        return getattr(self, item)


class CatalogDatasources(BaseModel, extra="forbid"):
    catalog_type: str = "scivision datasource catalog"
    name: str
    # Tuple: see comment on CatalogModelEntry
    entries: Tuple[CatalogDatasourceEntry, ...]

    @validator("entries")
    def name_unique_key(cls, entries):
        name_counts = Counter([entry['name'] for entry in entries])
        dups = [item for item, count in name_counts.items() if count > 1]

        if dups:
            raise ValueError(f"The 'name' field in the datasource catalog should be unique (duplicates: {dups})")

        return entries


def get_models():
    models_raw = pkgutil.get_data(__name__, "data/models.json")
    models = CatalogModels.parse_raw(models_raw)
    names = []
    for model_entry in models.entries:
        names.append(model_entry["name"])
    return names


modelEnumStrings = ((x, x) for x in get_models())
ModelEnum = Enum('ModelEnum', modelEnumStrings)


def get_datasources():
    datasources_raw = pkgutil.get_data(__name__, "data/datasources.json")
    datasources = CatalogDatasources.parse_raw(datasources_raw)
    names = []
    for datasources_entry in datasources.entries:
        names.append(datasources_entry["name"])
    return names


datasourceEnumStrings = ((x, x) for x in get_datasources())
DataEnum = Enum('DataEnum', datasourceEnumStrings)


class CatalogProjectEntry(BaseModel, extra="forbid", title="A project catalog entry"):
    # tasks, institution and tags are Tuples (rather than Lists) so
    # that they are immutable - Tuple is being used as an immutable
    # sequence here. This means that these fields are hashable, which
    # can be more convenient when included in a dataframe
    # (e.g. unique()). Could consider using a Frozenset for these
    # fields instead, since duplicates and ordering should not be
    # significant.
    name: str = Field(
        ...,
        title="Name",
        description="Short, unique name for the project (one or two words, "
        "under 20 characters recommended)",
    )
    header: str = Field(
        ...,
        title="Header",
        description="Header that will display at the top of the project page",
    )
    description: Optional[str] = Field(
        None,
        title="Description",
        description="Short description of the project (that will appear when hovering on the project thumbnail)",
    )
    page: str = Field(
        None,
        title="Page",
        description="Markdown formatted content for the project page",
    )
    models: Tuple[ModelEnum, ...] = Field(
        (),
        title="Models",
        description="Which models from the scivision catalog are used in the project?",
    )
    datasources: Tuple[DataEnum, ...] = Field(
        (),
        title="Datasources",
        description="Which datasources from the scivision catalog are used in the project?",
    )
    tasks: Tuple[TaskEnum, ...] = Field(
        (),
        title="Tasks",
        description="Which task (or tasks) do the CV models used in the project perform?",
    )
    institution: Tuple[str, ...] = Field(
        (),
        title="Institution(s)",
        description="A list of institutions that produced or are associated with "
        "the project (one per item)",
    )
    tags: Tuple[str, ...]

    def __getitem__(self, item):
        return getattr(self, item)


class CatalogProjects(BaseModel, extra="forbid"):
    catalog_type: str = "scivision project catalog"
    name: str
    # Tuple: see comment on CatalogProjectEntry
    entries: Tuple[CatalogProjectEntry, ...]

    @validator("entries")
    def name_unique_key(cls, entries):
        name_counts = Counter([entry['name'] for entry in entries])
        dups = [item for item, count in name_counts.items() if count > 1]

        if dups:
            raise ValueError(f"The 'name' field in the project catalog should be unique (duplicates: {dups})")

        return entries


def _coerce_datasources_catalog(
    datasources: Union[CatalogDatasources, os.PathLike, None]
) -> CatalogDatasources:
    """Returns a CatalogDatasources determined from the argument: either
    the one passed, or one loaded from a file
    """
    if isinstance(datasources, CatalogDatasources):
        return datasources
    elif isinstance(datasources, (bytes, str, os.PathLike)):
        datasources_raw = Path(datasources).read_text()
        return CatalogDatasources.parse_raw(datasources_raw)
    elif datasources is None:
        datasources_raw = pkgutil.get_data(__name__, "data/datasources.json")
        return CatalogDatasources.parse_raw(datasources_raw)
    else:
        raise TypeError("Cannot load datasource from unsupported type")


def _coerce_models_catalog(
    models: Union[CatalogModels, os.PathLike, None]
) -> CatalogModels:
    """Returns a CatalogModels determined from the argument: either the
    one passed, or one loaded from a file
    """
    if isinstance(models, CatalogModels):
        return models
    elif isinstance(models, (bytes, str, os.PathLike)):
        models_raw = Path(models).read_text()
        return CatalogModels.parse_raw(models_raw)
    elif models is None:
        models_raw = pkgutil.get_data(__name__, "data/models.json")
        return CatalogModels.parse_raw(models_raw)
    else:
        raise TypeError("Cannot load model from unsupported type")


def _coerce_projects_catalog(
    projects: Union[CatalogProjects, os.PathLike, None]
) -> CatalogProjects:
    """Returns a CatalogProjects determined from the argument: either the
    one passed, or one loaded from a file
    """
    if isinstance(projects, CatalogProjects):
        return projects
    elif isinstance(projects, (bytes, str, os.PathLike)):
        projects_raw = Path(projects).read_text()
        return CatalogProjects.parse_raw(projects_raw)
    elif projects is None:
        projects_raw = pkgutil.get_data(__name__, "data/projects.json")
        return CatalogProjects.parse_raw(projects_raw)
    else:
        raise TypeError("Cannot load project from unsupported type")


class QueryResult(ABC):
    @abstractmethod
    def to_dataframe(self) -> pd.DataFrame:
        ...

    def to_dict(self) -> Dict[str, Any]:
        return self.to_dataframe().to_dict(orient="records")


class PandasQueryResult(QueryResult):
    def __init__(self, result: pd.DataFrame):
        self._result = result

    def to_dataframe(self) -> pd.DataFrame:
        return self._result


class PandasCatalog:
    def __init__(self, datasources=None, models=None, projects=None):
        super().__init__()

        if isinstance(datasources, pd.DataFrame):
            self._datasources = datasources
        else:
            datasources_cat = _coerce_datasources_catalog(datasources)
            self._datasources = pd.DataFrame(
                [ent.dict() for ent in datasources_cat.entries]
            )

        if isinstance(models, pd.DataFrame):
            self._models = models
        else:
            models_cat = _coerce_models_catalog(models)
            self._models = pd.DataFrame([ent.dict() for ent in models_cat.entries])

        if isinstance(projects, pd.DataFrame):
            self._projects = projects
        else:
            projects_cat = _coerce_projects_catalog(projects)
            self._projects = pd.DataFrame([ent.dict() for ent in projects_cat.entries])

    @property
    def models(self) -> PandasQueryResult:
        return PandasQueryResult(self._models)

    @property
    def datasources(self) -> PandasQueryResult:
        return PandasQueryResult(self._datasources)

    @property
    def projects(self) -> PandasQueryResult:
        return PandasQueryResult(self._projects)

    def _compatible_models(self, datasource) -> PandasQueryResult:
        models_compatible_format = self._models[
            self._models.format == datasource["format"]
        ]

        models_compatible_format_labels = models_compatible_format[
            datasource["labels_provided"] | ~models_compatible_format.labels_required
        ]

        datasource_tasks = pd.DataFrame(datasource["tasks"], columns=["tasks"])
        models_compatible_tasks = (
            self._models[["name", "tasks"]]
            .explode("tasks")
            .merge(
                datasource_tasks,
                on="tasks",
                suffixes=("_model", "_datasource"),
            )
            .name.drop_duplicates()
        )
        result_df = models_compatible_format_labels[
            models_compatible_format_labels.name.isin(models_compatible_tasks)
        ]

        return PandasQueryResult(result_df)

    # Similar to _compatible_models, but for datasources.  Can't
    # cleanly combine these two functions, due to the asymmetry
    # between a model's 'labels_required', and datasource 'labels_provided'.
    # In particular, a model that doesn't require labels can still use
    # a datasource that provides them (if they are otherwise
    # compatible), but not vice versa.  For this reason, the distinct
    # names 'labels_provided' and 'labels_required' are used.
    def _compatible_datasources(self, model) -> PandasQueryResult:
        datasources_compatible_format = self._datasources[
            self._datasources.format == model["format"]
        ]

        datasources_compatible_format_labels = datasources_compatible_format[
            datasources_compatible_format.labels_provided | ~model["labels_required"]
        ]

        model_tasks = pd.DataFrame(model["tasks"], columns=["tasks"])
        datasources_compatible_tasks = (
            self._datasources[["name", "tasks"]]
            .explode("tasks")
            .merge(
                model_tasks,
                on="tasks",
                suffixes=("_model", "_datasource"),
            )
            .name.drop_duplicates()
        )
        result_df = datasources_compatible_format_labels[
            datasources_compatible_format_labels.name.isin(datasources_compatible_tasks)
        ]

        return PandasQueryResult(result_df)

    def compatible_models(self, datasource) -> PandasQueryResult:
        """Return all models that are compatible with datasource

        Parameters
        ----------
        datasource : str or dict-like
            Any dictionary-like (including CatalogDatasourceEntry) that
            has keys 'format', 'tasks' and 'labels_provided', representing
            these properties of the datasource.
            If a string is passed, this is used to look up the datasource
            (in `self._datasources`).

        Returns
        -------
        result: QueryResult
            A QueryResult instance containing the models compatible with the
            given datasource (convertible to a dict or pd.DataFrame).

        """
        if isinstance(datasource, str):
            return self._compatible_models(
                self._datasources.set_index("name").loc[datasource]
            )
        else:
            return self._compatible_models(datasource)

    def compatible_datasources(self, model) -> PandasQueryResult:
        """Return all datasources that are compatible with model

        Parameters
        ----------
        model : str or dict-like
            Any dictionary-like (including CatalogModelEntry) that has
            keys 'format', 'tasks' and 'labels_required', representing
            these properties of the model.
            If a string is passed, this is used to look up the model (in `self._models`).

        Returns
        -------
        result: QueryResult
            A QueryResult instance containing the datasources compatible with
            the given model (convertible to a dict or pd.DataFrame).

        """
        if isinstance(model, str):
            return self._compatible_datasources(
                self._models.set_index("name").loc[model]
            )
        else:
            return self._compatible_datasources(model)


default_catalog = PandasCatalog()



from scivision.io import package_from_config


def test_package_from_config(IMAGENET_MODEL_CONFIG):
    """Check that the install string for the imagenet example model looks as expected"""
    # see the url field in tests/test_model_scivision.yml:
    expected = 'git+https://github.com/alan-turing-institute/scivision-test-plugin@main#egg=scivision_test_plugin'
    assert package_from_config(IMAGENET_MODEL_CONFIG) == expected


def test_package_from_config_branch(IMAGENET_MODEL_CONFIG):
    """Check that the install string for the imagenet example model looks as expected"""
    # see the url field in tests/test_model_scivision.yml:
    expected = 'git+https://github.com/alan-turing-institute/scivision-test-plugin@test-branch#egg=scivision_test_plugin'
    assert package_from_config(IMAGENET_MODEL_CONFIG, branch='test-branch') == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import importlib
import subprocess
import sys


def _package_exists(config: dict) -> bool:
    """Check to see whether a package exists."""
    try:
        importlib.import_module(config["import"])
    except ModuleNotFoundError:
        return False

    return True


def package_from_config(config: dict, branch: str = "main") -> str:
    """Given a config return the pip install string."""
    install_str = config["url"]
    if install_str.endswith(".git"):
        install_str = install_str[:-4]
    install_branch = config.get("github_branch", branch)
    return f"git+{install_str}@{install_branch}#egg={config['import']}"


def _install(package, pip_install_args=None):
    """Install a package using pip."""

    if pip_install_args is None:
        pip_install_args = []

    subprocess.check_call(
        [sys.executable, "-m", "pip", "install", *pip_install_args, package]
    )


def install_package(
    config: dict,
    allow_install=False,  # allowed values: True, False, or the string "force"
    branch: str = "main",
):
    """Install the python package if it doesn't exist.

    Parameters
    ----------
    config: dict
      Dictionary of installation parameters for the package

    allow_install: one of True, False, or the string "force"
      - `True` means the package is allowed to call out to the system
        pip to install the package.
      - `False` means the package is still required, but rather than
        installing it, performs a check that it *already* exists
      - `"force"` installs the package, even if it exists already.
        This option is provided in case something gets stuck in an
        interactive session, but should otherwise be avoided if a
        reproducible environment is desired, since it will override
        any versions specified in an environment elsewhere, *including*
        any dependencies.

    branch: str
      If installing from a git repository, or a GitHub URL, use
      this branch
    """
    package = package_from_config(config, branch)
    exists = _package_exists(config)

    if allow_install == "force":
        # Install the package, whether it exists or not.

        # Note that the "--force-reinstall" argument to pip seems to
        # be inherited when installing dependencies, meaning that
        # these are force-upgraded even if a version satisfying the
        # requirements is present.

        # (--no-deps doesn't solve this, since the dependencies may
        # have changed between a version present and the force-updated
        # version)
        _install(package, pip_install_args=["--force-reinstall", "--no-cache-dir"])
    elif (allow_install and not exists):
        # The package is requested but isn't already installed: this
        # is the common case. The package and its dependencies will
        # be installed
        _install(package)
    elif not exists:
        # The package was requested, but allow install was false and
        # it isn't already installed: raise a runtime error
        raise Exception(
            "Package does not exist. Try installing it with: \n"
            f"`!pip install -e {package}`"
        )



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

from IPython.display import display
from distinctipy.distinctipy import get_colors, get_text_color
from matplotlib.image import AxesImage
from matplotlib.colors import rgb2hex
import numpy as np
import pandas as pd
from PIL import Image, ImageDraw, ImageFont


def display_objects(image: np.ndarray,
                    predictions: list,
                    label_nums: bool = False) -> AxesImage:
    """Display an image with colored bounding boxes for objects identified by a
    scivision object detection model.

    Parameters
    ----------
    image : np.ndarray
        A numpy ndarray representation of an image.
    predictions : list
        A list of dictionaries, on for each detected object containing 'score',
        'label' and 'box' keys.
        'score' should be a float between 0 and 1.
        'label' should be a string that labels the object.
        'box' should be a dict with 'xmin', 'ymin', 'xmax' and 'ymax' keys,
        which denote the boundaries of each box (detected object).
    label_nums : bool = False
        When True, bounding boxes are numbered in addition to being colored.

    Returns
    -------
    The input image with colored bounding boxes and an accompanying pandas
    dataframe legend showing corresponding labels and scores for each object.
    """
    # Convert image to RGB
    pillow_image = Image.fromarray(image.to_numpy(), 'RGB')

    # Load font needed for numbering of bounding boxes and set size
    font = ImageFont.load_default()

    # Generate visually distinct colors for each object bounding box
    # We want a list of hexadecimal colors to use, and distinctipy can also
    # give us a matching list of the most appropriate text colors
    # for these colors (black or white)
    rgb_colors = get_colors(len(predictions), colorblind_type="Deuteranomaly")
    hex_colors = []
    text_hex_colors = []
    for color in rgb_colors:
        hex_colors.append(rgb2hex(color))
        text_hex_colors.append(rgb2hex(get_text_color(color)))

    # Place the object bounding boxes onto the image and use IPython to display
    index = 0
    for bounding_box in predictions:
        box = bounding_box["box"]
        im_with_rectangle = ImageDraw.Draw(pillow_image)
        im_with_rectangle.rounded_rectangle((box["xmin"], box["ymin"],
                                             box["xmax"], box["ymax"]),
                                            outline=hex_colors[index],
                                            width=2, radius=7)
        if label_nums:
            im_with_rectangle.text((box["xmin"] + 4, box["ymin"] + 3),
                                   str(index), fill=hex_colors[index],
                                   font=font)
        index += 1
    display(pillow_image)

    # Return a colored pandas df table relating to the objects shown in image
    object_predictions = pd.DataFrame(predictions).drop('box', 1)
    object_predictions['bbox'] = object_predictions.index

    def get_col(s):
        """Func to color pandas df"""
        col = 'background-color: ' + hex_colors[s.bbox]
        col += '; color: ' + text_hex_colors[s.bbox]
        return [col] * 3

    return object_predictions.style.apply(get_col, axis=1)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

import os
from urllib.parse import urlparse

import fsspec
import intake
import yaml

from ..koala import koala
from .installer import install_package
from .wrapper import PretrainedModel, Datasource

import warnings
import xarray


def _is_url(path: os.PathLike) -> bool:
    return urlparse(path).scheme in (
        "http",
        "https",
    )


def _parse_url(path: os.PathLike, branch: str = "main"):
    """Parse a URL and convert to a raw github url if necessary."""

    parsed = urlparse(path)

    if parsed.netloc not in ["github.com"]:
        return parsed.geturl()

    if parsed.netloc == "github.com":
        # construct the new github path
        parsed = parsed._replace(netloc="raw.githubusercontent.com")
        split = list(filter(None, parsed.path.split("/")))
        if branch not in split:
            new_path = "/".join(split[:2]) + f"/{branch}/" + "/".join(split[2:])
        else:
            new_path = "/".join(split)
        new_path = new_path.replace("/blob", "")

        parsed = parsed._replace(path=new_path)
        return parsed.geturl()
    else:
        raise NotImplementedError


def _get_model_configs(
    full_config: dict, load_multiple: bool = False, model: str = "default"
):
    """Get one config per model from a multi-model config.

    Parameters
    ----------
    full_config : dict
        Dictionary of a .scivision/model.yml config loaded from yaml.
    load_multiple : bool, default = False
        Modifies the return to be a list of scivision.PretrainedModel's.
    model : str, default = default
        Specify the name of a model to get the config for.

    Returns
    -------
    List of dictionaries
        One dictionary per config model.
    """
    # Create a list that will contain one or multiple model configs
    config_list = []
    if (
        "models" in full_config
    ):  # if there are multiple models specified in the config yml
        if load_multiple:
            # Create a config for each model
            for model_dict in full_config["models"]:
                new_config = {}
                new_config["name"] = full_config["name"]
                new_config["url"] = full_config["url"]
                new_config["import"] = full_config["import"]
                new_config["model"] = model_dict["model"]
                new_config["args"] = model_dict["args"]
                new_config["prediction_fn"] = model_dict["prediction_fn"]
                config_list.append(new_config)
        else:
            # Choose the first model in the list by default
            if model == "default":
                full_config["model"] = full_config["models"][0]["model"]
                full_config["args"] = full_config["models"][0]["args"]
                full_config["prediction_fn"] = full_config["models"][0]["prediction_fn"]
            # Choose the named model:
            else:
                for model_dict in full_config["models"]:
                    if model_dict["model"] == model:
                        full_config["model"] = model_dict["model"]
                        full_config["args"] = model_dict["args"]
                        full_config["prediction_fn"] = model_dict["prediction_fn"]
                        break
                # Check that a model of name "model" in .scivision/model.yml config
                if "model" not in full_config:
                    raise ValueError(
                        "model of name " + model + " not found in config yaml"
                    )
            config_list.append(full_config)
    else:  # if there is a single model specified in the config yml
        if load_multiple:
            warnings.warn(
                "Only one model found in config yaml "
                "(i.e., no 'models' section in the config file), "
                "will load that one..."
            )
        # Check that a model of name "model" in .scivision/model.yml config
        if model != "default" and full_config["model"] != model:
            raise ValueError("model of name " + model + " not found in config yaml")
        config_list.append(full_config)
    return config_list


@koala
def load_pretrained_model(
    path: os.PathLike,
    branch: str = "main",
    allow_install: bool = False,
    model_selection: str = "default",
    load_multiple: bool = False,
    *args,
    **kwargs,
) -> PretrainedModel:
    """Load a pre-trained model.

    Parameters
    ----------
    path : PathLike
        The filename, path or URL of a pretrained model description.
    branch : str, default = main
        Specify the name of a github branch if loading from github.
    allow_install : bool, default = False
        Allow installation of remote package via pip.
    model_selection : str, default = default
        Specify the name of the model if there is > 1 in the model repo package.
    load_multiple : bool, default = False
        Modifies the return to be a list of scivision.PretrainedModel's.

    Returns
    -------
    pretrained_model : scivision.PretrainedModel
        The instantiated pre-trained model.
    """

    if _is_url(path):
        path = _parse_url(path, branch)
    # check that this is a path to a yaml file
    # if not, assume it is a repo containing ".scivision/model.yml"
    if not path.endswith(
        (
            ".yml",
            ".yaml",
        )
    ):
        path = path + ".scivision/model.yml"
    # fsspec will throw an error if the path does not exist
    file = fsspec.open(path)
    # parse the config file:
    with file as config_file:
        stream = config_file.read()
        config = yaml.safe_load(stream)
    config_list = _get_model_configs(config, load_multiple, model_selection)
    loaded_models = []
    for config in config_list:
        # make sure a model at least has an input to the function
        assert "X" in config["prediction_fn"]["args"].keys()

        # try to install the package if necessary
        install_package(config, allow_install=allow_install, branch=branch)

        loaded_models.append(PretrainedModel(config))
    if load_multiple:
        return loaded_models
    # By default, return a single PretrainedModel
    return loaded_models[0]


def load_dataset(
    path: os.PathLike, branch: str = "main"
) -> intake.catalog.local.YAMLFileCatalog:
    """Load a dataset.

    Parameters
    ----------
    path : PathLike
        The filename, path or URL of an intake catalog, which links to a dataset.
    branch : str, default = main
        Specify the name of a github branch if loading from github.

    Returns
    -------
    intake.catalog.local.YAMLFileCatalog
        An intake catalog object representing the loaded dataset (see `intake.readthedocs <https://intake.readthedocs.io/en/latest/api_user.html?highlight=catalog.local.YAMLFileCatalog#intake.catalog.local.YAMLFileCatalog>`_).
    """

    if _is_url(path):
        path = _parse_url(path, branch)

    # check that this is a path to a yaml file
    # if not, assume it is a repo containing ".scivision/data.yml"
    if not path.endswith(
        (
            ".yml",
            ".yaml",
        )
    ):
        path = path + ".scivision/data.yml"

    # fsspec will throw an error if the path does not exist
    file = fsspec.open(path)

    # parse the config file to see if it's a data plugin:
    with file as config_file:
        stream = config_file.read()
        config = yaml.safe_load(stream)
    if "import" in config:
        return load_data_from_plugin(config, branch)

    # if not a data plugin, assume an intake config
    intake_cat = intake.open_catalog(path)

    return intake_cat


def load_data_from_plugin(
    config: dict, branch: str = "main"
) -> xarray.Dataset:
    """Load a dataset from a data plugin (python package).

    Parameters
    ----------
    config : dict
        The loaded data.yml config.
    branch : str, default = main
        Specify the name of a github branch if loading from github.

    Returns
    -------
    xarray.Dataset
        The dataset to be visualised, loaded via xarray.
    """

    # install the package
    install_package(config, allow_install=True, branch=branch)

    return Datasource(config)



from scivision.io import PretrainedModel
from skimage.io import imread


def test_pretrained_model(IMAGENET_MODEL):
    """Check that a pretrained model object can be created from config"""
    assert isinstance(IMAGENET_MODEL, PretrainedModel)


def test_pretrained_model_predict(KOALA, IMAGENET_MODEL):
    """Check that an example pretrained model's predict function returns expected value"""
    prediction = IMAGENET_MODEL.predict(imread(KOALA))
    assert 'koala' in prediction  # expects a string that also includes confidence



from scivision.koala import get_koala_log, koala


@koala
def _dummy_fn(x, *, Y: str = "test"):
    """A dummy function."""
    return


def test_koala_logging():
    """Test that koala logging is functional by executing a dummy function
    decorated with the `@koala` logging decorator."""
    log_old = get_koala_log()
    _ = _dummy_fn(1, Y="test")
    log_new = get_koala_log()
    assert len(log_new) > len(log_old)
